use color_eyre::Result;
use ratatui::{
    layout::{Alignment, Rect},
    text::Line,
    widgets::{Block, Padding, Paragraph, Wrap},
    Frame,
};

use super::Component;
use crate::{
    action::Action, 
    palette::StylePalette, 
    types::EraInfo, 
    widgets::{PixelSize, BigText},
};

#[derive(Debug, Default)]
pub struct CurrentEra{
    era: EraInfo,
    palette: StylePalette,
}

impl CurrentEra {
    const ERA_OFFSET_IN_SLOTS:  u64 = 2_400 * 6;
    const ERA_OFFSET_IN_MILLIS: u64 = Self::ERA_OFFSET_IN_SLOTS * 6_000;

    const MILLIS_IN_DAY:    u64 = 86_400_000;
    const MILLIS_IN_HOUR:   u64 = 3_600_000;
    const MILLIS_IN_MINUTE: u64 = 60_000;

    fn update_era(&mut self, era_info: EraInfo) -> Result<()> {
        self.era = era_info;
        Ok(())
    }
}

impl Component for CurrentEra {
    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetActiveEra(era_info) => self.update_era(era_info)?,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, place] = super::explorer_era_info_layout(area);

        let current_time = std::time::SystemTime::now()
            .duration_since(std::time::UNIX_EPOCH)
            .expect("time went backwards")
            .as_millis() as u64;

        let era_start = self.era.start.unwrap_or_default();
        let era_end = era_start + Self::ERA_OFFSET_IN_MILLIS;

        let (difference, reversed_char) = if era_end > current_time {
            (era_end - current_time, "")
        } else {
            (current_time - era_end, "< ")
        };

        let hours = (difference % Self::MILLIS_IN_DAY) / Self::MILLIS_IN_HOUR;
        let minutes = (difference % Self::MILLIS_IN_HOUR) / Self::MILLIS_IN_MINUTE;
        let seconds = difference % Self::MILLIS_IN_MINUTE / 1000;

        let text = self.era.index.to_string();
        let big_time = hours > 0;
            
        let (border_style, border_type) = self.palette.create_border_style(false);

        let height = place.as_size().height;
        let width = place.as_size().width;
        let text_width = text.len() as u16 * 4 + 2;

        if width < text_width || height < 7 {
            let text = vec![
                Line::from(text),
                Line::from(format!("{}{}{} {}{}",
                        reversed_char,
                        if big_time { hours } else { minutes },
                        if big_time { "hrs" } else { "mins" },
                        if big_time { minutes } else { seconds },
                        if big_time { "mins" } else { "secs" })),
            ];
            let paragraph = Paragraph::new(text)
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_alignment(Alignment::Right)
                    .title_style(self.palette.create_title_style())
                    .padding(Padding::new(0, 0, (height.saturating_sub(3)) / 2, 0))
                    .title("Era"))
                .alignment(Alignment::Center)
                .wrap(Wrap { trim: true });
            frame.render_widget(paragraph, place);
        } else {
            let big_text = BigText::builder()
                .centered()
                .pixel_size(PixelSize::Quadrant)
                .style(self.palette.create_text_style(false))
                .lines(vec![
                    text.into(),
                ])
                .build();
            let paragraph = Paragraph::new("")
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_style(self.palette.create_title_style())
                    .title_top(Line::from("Era").right_aligned())
                    .title_top(Line::from(format!("{}{} {}{}",
                                if big_time { hours } else { minutes },
                                if big_time { "hrs" } else { "mins" },
                                if big_time { minutes } else { seconds },
                                if big_time { "mins" } else { "secs" }))
                        .centered()))
                .alignment(Alignment::Center)
                .wrap(Wrap { trim: true });
            frame.render_widget(paragraph, place);

            let height_offset = height.saturating_sub(2) / 2;
            let place = place.offset(ratatui::layout::Offset { x: 1, y: height_offset as i32 })
                .intersection(place);
            frame.render_widget(big_text, place);
        }
        Ok(())
    }
}
