use std::collections::VecDeque;
use color_eyre::Result;
use ratatui::{
    layout::{Alignment, Rect}, 
    text::Line, 
    widgets::{Bar, BarChart, BarGroup, Block},
    Frame
};

use super::Component;
use crate::action::Action;

#[derive(Debug, Default)]
pub struct ExtrinsicsChart {
    extrinsics: VecDeque<(u32, usize)>,
}

impl ExtrinsicsChart {
    const MAX_LEN: usize = 50;
    const BAR_WIDTH: usize = 6;
    const BAR_GAP: usize = 1;

    fn extrinsics_bar_chart(&self, width: u16) -> BarChart {
        let length = (width as usize) / (Self::BAR_WIDTH + Self::BAR_GAP);

        let bars: Vec<Bar> = self.extrinsics
            .iter()
            .rev()
            .take(length)
            .map(|(block_number, ext_len)| self.extrinsic_vertical_bar(block_number, ext_len))
            .collect();

        BarChart::default()
            .data(BarGroup::default().bars(&bars))
            .block(Block::bordered().title_alignment(Alignment::Right).title("Tx. Heat Map"))
            .bar_width(8)
            .bar_gap(1)
    }

    fn extrinsic_vertical_bar(&self, block_number: &u32, ext_len: &usize) -> Bar {
        Bar::default()
            .value(*ext_len as u64)
            .label(Line::from(format!("..{}", block_number % 100)))
            .text_value(ext_len.to_string())
    }

    fn update_extrinsics_length(&mut self, block_number: u32, extrinsics_length: usize) -> Result<()> {
        match self.extrinsics.back() {
            Some(back) => {
                if back.0 < block_number {
                    self.extrinsics.push_back((block_number, extrinsics_length));
                    if self.extrinsics.len() > Self::MAX_LEN {
                        let _ = self.extrinsics.pop_front();
                    }
                }
            },
            None => {
                self.extrinsics.push_back((block_number, extrinsics_length));
            }
        };

        Ok(())
    }
}

impl Component for ExtrinsicsChart {
    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::ExtrinsicsLength(block, length) => self.update_extrinsics_length(block, length)?,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, place] = super::explorer_header_layout(area);
        frame.render_widget(self.extrinsics_bar_chart(place.as_size().width), place);
        Ok(())
    }
}
