use color_eyre::Result;
use crossterm::event::KeyEvent;
use ratatui::{
    layout::{Constraint, Flex, Layout, Rect},
    Frame,
};

use super::Component;
use crate::action::Action;

mod latest_block;
mod finalized_block;
mod block_ticker;
mod current_era;
mod current_epoch;
mod extrinsics_chart;
mod explorer_blocks;

use latest_block::LatestBlock;
use finalized_block::FinalizedBlock;
use block_ticker::BlockTicker;
use current_era::CurrentEra;
use current_epoch::CurrentEpoch;
use extrinsics_chart::ExtrinsicsChart;
use explorer_blocks::ExplorerBlocks;

pub struct Explorer {
    components: Vec<Box<dyn Component>>
}

impl Default for Explorer {
    fn default() -> Self {
        Self {
            components: vec![
                Box::new(BlockTicker::default()),
                Box::new(LatestBlock::default()),
                Box::new(FinalizedBlock::default()),
                Box::new(CurrentEra::default()),
                Box::new(CurrentEpoch::default()),
                Box::new(ExtrinsicsChart::default()),
                Box::new(ExplorerBlocks::default()),
            ]
        }
    }
}

impl Component for Explorer {
    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        for component in self.components.iter_mut() {
            component.handle_key_event(key)?;
        }
        Ok(None)
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        for component in self.components.iter_mut() {
            component.update(action.clone())?;
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let screen = super::screen_layout(area);
        for component in self.components.iter_mut() {
            component.draw(frame, screen)?;
        }
        Ok(())
    }
}

pub fn explorer_layout(area: Rect) -> [Rect; 3] {
    Layout::vertical([
        Constraint::Percentage(30),
        Constraint::Percentage(40),
        Constraint::Percentage(30),
    ]).areas(area)
}

pub fn explorer_header_layout(area: Rect) -> [Rect; 2] {
    let [header, _, _] = explorer_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).areas(header)
}

pub fn explorer_info_layout(area: Rect) -> [Rect; 2] {
    let [info, _]  = explorer_header_layout(area);
    Layout::vertical([
        Constraint::Percentage(100),
        Constraint::Percentage(100),
    ]).areas(info)
}

pub fn explorer_block_info_layout(area: Rect) -> [Rect; 3] {
    let [blocks, _]  = explorer_info_layout(area);
    Layout::horizontal([
        Constraint::Percentage(33),
        Constraint::Percentage(33),
        Constraint::Percentage(33),
    ]).flex(Flex::SpaceBetween).areas(blocks)
}

pub fn explorer_era_info_layout(area: Rect) -> [Rect; 2] {
    let [_, blocks]  = explorer_info_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).flex(Flex::SpaceBetween).areas(blocks)
}

pub fn explorer_scrollbars_layout(area: Rect) -> [Rect; 2] {
    let [_, place, _] = explorer_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).areas(place)
}
