use color_eyre::Result;
use ratatui::{prelude::*, widgets::*};
use tokio::sync::mpsc::UnboundedSender;
use crossterm::event::{KeyEvent, KeyCode};

use super::Component;
use super::palette::StylePalette;
use crate::{action::Action, app::Mode};

pub struct Menu {
    command_tx: Option<UnboundedSender<Action>>,
    items: Vec<String>,
    current_item_index: usize,
    is_active: bool,
    palette: StylePalette,
}

impl Default for Menu {
    fn default() -> Self {
        Menu::new()
    }
}

impl Menu {
    pub fn new() -> Self {
        Self {
            command_tx: None,
            items: vec![
                String::from("Explorer"),
                String::from("Wallet"),
                String::from("Prices"),
                String::from("Staking"),
                String::from("Governance"),
                String::from("Operations"),
            ],
            current_item_index: Default::default(),
            is_active: true,
            palette: StylePalette::default(),

        }
    }

    fn move_current_up(&mut self) -> Result<()> {
        self.current_item_index = self
            .current_item_index
            .saturating_sub(1);

        if let Some(command_tx) = &self.command_tx {
            match self.current_item_index {
                0 => command_tx.send(Action::SetMode(Mode::Explorer))?,
                _ => command_tx.send(Action::SetMode(Mode::Empty))?,
            }
        };
        Ok(())
    }

    fn move_current_down(&mut self) -> Result<()> {
        let new_current = self.current_item_index + 1;
        if new_current < self.items.len() {
            self.current_item_index = new_current;
        }

        if let Some(command_tx) = &self.command_tx {
            match self.current_item_index {
                0 => command_tx.send(Action::SetMode(Mode::Explorer))?,
                _ => command_tx.send(Action::SetMode(Mode::Empty))?,
            }
        };
        Ok(())
    }

    fn set_active(&mut self) -> Result<()> {
        self.is_active = true;
        if let Some(command_tx) = &self.command_tx {
            match self.current_item_index {
                0 => command_tx.send(Action::SetMode(Mode::Explorer))?,
                _ => command_tx.send(Action::SetMode(Mode::Empty))?,
            }
        };
        Ok(())
    }

    fn unset_active(&mut self) -> Result<()> {
        self.is_active = false;
        if let Some(command_tx) = &self.command_tx {
            match self.current_item_index {
                0 => command_tx.send(Action::SetMode(Mode::ExplorerActive))?,
                _ => command_tx.send(Action::SetMode(Mode::EmptyActive))?,
            }
        };
        Ok(())
    }
}

impl Component for Menu {
    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        self.command_tx = Some(tx);
        Ok(())
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        match key.code {
           KeyCode::Up | KeyCode::Char('k') if self.is_active => self.move_current_up()?, 
           KeyCode::Down | KeyCode::Char('j') if self.is_active => self.move_current_down()?, 
           KeyCode::Enter | KeyCode::Char('l') | KeyCode::Right if self.is_active => self.unset_active()?,
           KeyCode::Esc if !self.is_active => self.set_active()?,
           _ => {},
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [menu, _] = super::menu_layout(area);
        let width = menu.as_size().width as usize;

        let mut text = vec![];
        for (position, item) in self.items.iter().enumerate() {
            let active = position == self.current_item_index;
            let line = Line::raw(format!("{:^width$}", item))
                .style(self.palette.create_text_style(active))
                .centered();
            text.push(line);
        }

        let (color, border_type) = self.palette.create_border_style(self.is_active);
        let block = Block::bordered()
            .border_style(color)
            .border_type(border_type);

        let paragraph = Paragraph::new(text)
            .block(block)
            .alignment(Alignment::Center);

        frame.render_widget(paragraph, menu);
        Ok(())
    }
}
